import string
import math
from itertools import product

from .fields import PathValue


class PathGenerator(object):
    def __init__(self, prefix=None, skip=None):
        combinations = string.digits + string.ascii_letters

        self.skip_paths = [] if skip is None else skip[:]
        self.path_prefix = prefix if prefix else []
        self.product_iterator = product(
            combinations,
            repeat=self.guess_the_label_size(
                path_size=len(self.skip_paths), combination_size=len(combinations)
            ),
        )

    def __iter__(self):
        return self

    def __next__(self):
        for val in self.product_iterator:
            label = "".join(val)
            path = PathValue(self.path_prefix + [label])
            if path not in self.skip_paths:
                return path

    next = __next__

    @staticmethod
    def guess_the_label_size(path_size: int, combination_size: int) -> int:
        # The theoritical limit for this at the time of writing is 2_538_557_185_841_324_496 (python 3.12.2)
        calculated_path_size = -1  # -1 is here for 0th index items
        # The theoritical limit for this at the time of writing is 32 (python 3.12.2)
        label_size = 0

        # THIS IS AN VERY IMPORTANT CHECK
        last = 0

        while True:
            possible_cominations = math.comb(combination_size, label_size)
            if last > possible_cominations:
                raise ValueError("We approached the limit of `math.comb`")

            last = possible_cominations
            calculated_path_size += possible_cominations

            if calculated_path_size > path_size and label_size != 0:
                break

            label_size += 1

        return label_size





from django.apps import AppConfig


class DjangoLtreeConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_ltree"

    def ready(self):
        from . import checks as checks
        from . import lookups as lookups
        from . import functions as functions



from django.db import models

from django_ltree.paths import PathGenerator


class TreeQuerySet(models.QuerySet):
    def roots(self):
        return self.filter(path__depth=1)

    def children(self, path):
        return self.filter(path__descendants=path, path__depth=len(path) + 1)


class TreeManager(models.Manager):
    def get_queryset(self):
        return TreeQuerySet(model=self.model, using=self._db).order_by("path")

    def roots(self):
        return self.filter().roots()

    def children(self, path):
        return self.filter().children(path)

    def create_child(self, parent=None, **kwargs):
        paths_in_use = parent.children() if parent else self.roots()
        prefix = parent.path if parent else None
        path_generator = PathGenerator(
            prefix,
            skip=paths_in_use.values_list("path", flat=True),
        )
        kwargs["path"] = path_generator.next()
        return self.create(**kwargs)



from django.db.models import Lookup

from .fields import PathField


class SimpleLookup(Lookup):
    lookup_operator = "="  # type: str

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return f"{lhs} {self.lookup_operator} {rhs}", [
            *lhs_params,
            *rhs_params,
        ]


@PathField.register_lookup
class EqualLookup(Lookup):
    lookup_name = "exact"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return f"{lhs} = {rhs}", [*lhs_params, *rhs_params]


@PathField.register_lookup
class AncestorLookup(SimpleLookup):
    lookup_name = "ancestors"
    lookup_operator = "@>"


@PathField.register_lookup
class DescendantLookup(SimpleLookup):
    lookup_name = "descendants"
    lookup_operator = "<@"


@PathField.register_lookup
class MatchLookup(SimpleLookup):
    lookup_name = "match"
    lookup_operator = "~"


@PathField.register_lookup
class ContainsLookup(SimpleLookup):
    lookup_name = "contains"
    lookup_operator = "?"



from django.core.checks import Warning, register


@register
def check_database_backend_is_postgres(app_configs, **kwargs):
    from django.conf import settings

    errors = []
    valid_dbs = ["postgres", "postgis"]

    if "default" in settings.DATABASES and all(
        d not in settings.DATABASES["default"]["ENGINE"] for d in valid_dbs
    ):
        errors.append(
            Warning(
                "django_ltree needs postgres to support install the ltree extension.",
                hint="Use the postgres engine or ignore if you already use a custom engine for postgres",
            )
        )

    return errors
